from textual.screen import ModalScreen
from textual.widgets import Label,Button
from textual.containers import Grid

CSS ='''

NetworkErrorScreen {
    align: center middle;
}

#network-error-dialog-box{
    grid-size: 2 3;
    grid-gutter: 1 2;
    grid-rows: 1fr 2fr 3;
    padding: 0 1;
    width: 50;
    height: 11;
    border: thick $background 80%;
    background: $surface;
}

#error-label {
    column-span: 2;
    height: 1fr;
    width: 2fr;
    content-align: center middle;
}

#error-heading {
    column-span: 2;
    height: 1fr;
    width: 1fr;
    content-align: center middle;
    color:$primary
}

#retry-button {
    width: 100%;
    column-span: 1;
}

#exit-button {
    width: 100%;
    column-span: 1;
}

'''

class NetworkErrorScreen(ModalScreen):
    DEFAULT_CSS = CSS
    def compose(self):
        yield Grid(
            Label("[bold]NETWORK ERROR",id="error-heading"),
            Label("CHECK YOUR INTERNET CONNECTION", id="error-label"),
            Button("RETRY", id="retry-button",variant='warning'),
            Button("EXIT", variant="error", id="exit-button"),
            id="network-error-dialog-box",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "exit-button":
            exit()
        else:
            self.dismiss("retry")


from spotidex.src.utils import *
from spotidex.src.static import *

import re, logging
from pathlib import Path

from tempfile import TemporaryDirectory
import shutil
import yt_dlp.YoutubeDL
from spotipy import Spotify
from spotipy.oauth2 import SpotifyClientCredentials

class SpotifyDownloader:
    def __init__(self) -> None:
        self.spotidex_path = get_spotidex_data_directory()
        self.ffmpeg_path = get_ffmpeg()
        self.is_library = False
        self.progress_dict = {}
        self.total_tracks = 1
   

        client_cred_manager = SpotifyClientCredentials(
            client_id,
            client_secret,
        )
        self.sp = Spotify(client_credentials_manager=client_cred_manager)
        

    def download_track(
        self,
        link,
        custom_hook=None,
        path=None,
        quality="high",
        metadata=True,
    ):
        logging.getLogger("pytube").setLevel(logging.ERROR)

        data = self.sp.track(link.split("/")[-1].split("?")[0])
        track_artist = ",".join([artist["name"] for artist in data["artists"]])
        track_name = data["name"]
        query = f"{track_name} song by {track_artist} official lyrics "

        file_name = re.sub(r"[^\w\d(),.\- ]", "", data["name"])
        download_path = Path.cwd() if path == None else path
        for i in range(1,1000000):
            if Path(download_path,file_name+".mp3").exists():
                file_name = file_name.split(" (")[0]+f" ({i})"
            else:
                break
        if self.is_library:
            options = {
            "format": "bestaudio/best",
            "outtmpl": str(download_path)+'\\'+file_name,
            "quiet": True,
            "noprogress": True,
            "progress": "false",
            "ffmpeg_location": str(self.ffmpeg_path),
            "postprocessors": [
                {
                    "key": "FFmpegExtractAudio",
                    "preferredcodec": "mp3",
                    "preferredquality": bitrate[quality],
                }
            ],
            "retries": 20,
            "progress_hooks": [lambda d: progress_hook(self, d, custom_hook)]
            if custom_hook != None
            else [],
        }

            ydl = yt_dlp.YoutubeDL(options)
            yt_video_result = ydl.extract_info(f"ytsearch:{query}", download=False)
            first_yt_video_link = yt_video_result["entries"][0]["webpage_url"]
            yt_video_download = ydl.extract_info(first_yt_video_link, download=True)
            downloaded_path = ydl.prepare_filename(yt_video_download)

            if metadata:
                add_metadata(data, Path(downloaded_path + ".mp3").resolve())
        else:
            with TemporaryDirectory(dir=str(self.spotidex_path)) as temp_folder:

                options = {
                    "format": "bestaudio/best",
                    "outtmpl": str(Path(temp_folder))+'\\'+file_name,
                    "quiet": True,
                    "noprogress": True,
                    "progress": "false",
                    "ffmpeg_location": str(self.ffmpeg_path),
                    "postprocessors": [
                        {
                            "key": "FFmpegExtractAudio",
                            "preferredcodec": "mp3",
                            "preferredquality": bitrate[quality],
                        }
                    ],
                    "retries": 20,
                    "progress_hooks": [lambda d: progress_hook(self, d, custom_hook)]
                    if custom_hook != None
                    else [],
                }

                ydl = yt_dlp.YoutubeDL(options)
                yt_video_result = ydl.extract_info(f"ytsearch:{query}", download=False)
                first_yt_video_link = yt_video_result["entries"][0]["webpage_url"]
                yt_video_download = ydl.extract_info(first_yt_video_link, download=True)
                downloaded_path = ydl.prepare_filename(yt_video_download)

                if metadata:
                    add_metadata(data, Path(downloaded_path + ".mp3").resolve())

                if not self.is_library and custom_hook is not None:
                    progress_hook(self, {"status": "downloaded"}, custom_hook)

                if not self.is_library:
                    for file_path in Path(temp_folder).iterdir():
                        shutil.move(file_path, download_path)

                    if custom_hook is not None:
                        progress_hook(self, {"status": "transfered"}, custom_hook)
        
        

    def download_playlist(
        self,
        link,
        custom_hook=None,
        download_path=Path.cwd(),
        quality="high",
        metadata=True,
        make_folder=True,
    ):
        playlist_id = link.split("/")[-1].split("?")[0]
        playlist_data = self.sp.playlist(playlist_id)
        track_links = extract_track_links(self.sp, link)

        self.is_library = True
        self.total_tracks = len(track_links)

        folder_name = re.sub(r"[^\w\d(),.\- ]", "", playlist_data["name"])
        for i in range(1,1000000):
            if Path(download_path,folder_name).exists():
                folder_name = folder_name.split(" (")[0]+f" ({i})"
            else:
                break
        with TemporaryDirectory(dir=str(self.spotidex_path)) as temp_folder:
            if make_folder == True:
                temp_path = Path(temp_folder).joinpath(folder_name)
            else:
                temp_path = Path(temp_folder)   

            done,not_done = pool_download(
                    self,
                    list(set(track_links)),
                    custom_hook,
                    temp_path,
                    quality,
                    metadata,
                )
            if len(not_done) != 0:
                raise ConnectionError
            
            if custom_hook != None:
                progress_hook(self, {"status": "downloaded"}, custom_hook)

            for file_path in Path(temp_folder).iterdir():
                shutil.move(str(file_path), str(download_path))

            if custom_hook != None:
                progress_hook(self, {"status": "transfered"}, custom_hook)
            
        


    def download_album(
        self,
        link,
        custom_hook=None,
        download_path=Path.cwd(),
        quality="high",
        metadata=True,
        make_folder=True,
    ):
        album_id = link.split("/")[-1].split("?")[0]
        album_data = self.sp.album(album_id)
        track_links = extract_track_links(self.sp, link)

        self.is_library = True
        self.total_tracks = len(track_links)
        
        folder_name,i = re.sub(r"[^\w\d(),.\-\[\] ]", "", folder_name),1
        while Path(download_path, folder_name).exists():
            folder_name,i = folder_name.split(" (")[0] + f" ({i})",i+1
        folder_name = re.sub(r"[^\w\d(),.\- ]", "", album_data["name"])
        for i in range(1,1000000):
            if Path(download_path,folder_name).exists():
                folder_name = folder_name.split(" (")[0]+f" ({i})"
            else:
                break
        with TemporaryDirectory(dir=str(self.spotidex_path)) as temp_folder:
            if make_folder == True:
                temp_path = Path(temp_folder).joinpath(folder_name)
            else:
                temp_path = Path(temp_folder)   

            done,not_done = pool_download(
                    self,
                    list(set(track_links)),
                    custom_hook,
                    temp_path,
                    quality,
                    metadata,
                )
            if len(not_done) != 0:
                raise ConnectionError

            progress_hook(self, {"status": "downloaded"}, custom_hook)
            with self.temp_folder:
                for file_path in Path(temp_folder).iterdir():
                    shutil.move(str(file_path), str(download_path))
            progress_hook(self, {"status": "transfered"}, custom_hook)
        


    def download_multiple_songs(
        self,
        links,
        folder_name,
        custom_hook=None,
        download_path=Path.cwd(),
        quality="high",
        metadata=True,
        make_folder=True,
    ):
        self.is_library = True
        self.total_tracks = len(links)
        with TemporaryDirectory(dir=str(self.spotidex_path)) as temp_folder:
            if make_folder == True:
                folder_name,i = re.sub(r"[^\w\d(),.\-\[\] ]", "", folder_name),1
                while Path(download_path, folder_name).exists():
                    folder_name,i = folder_name.split(" (")[0] + f" ({i})",i+1
                temp_path = Path(temp_folder).joinpath(folder_name)
            else:
                temp_path = Path(temp_folder)   


            done,not_done = pool_download(
                    self, list(set(links)), custom_hook, temp_path, quality, metadata
                )
            if len(not_done) != 0:
                raise ConnectionError
            

            progress_hook(self, {"status": "downloaded"}, custom_hook)
            for file_path in Path(temp_folder).iterdir():
                shutil.move(str(file_path), str(download_path))
            progress_hook(self, {"status": "transfered"}, custom_hook)
        
    



import requests
from mutagen.id3 import ID3, APIC, TPE1, TIT2, TALB

from rich.panel import Panel
from rich.console import Console
from rich.table import Table
import os
import stat
import platform
from pathlib import Path
from spotidex.src.static import FFMPEG_URLS
from tqdm import tqdm
import concurrent.futures
import subprocess

console = Console()
clear_screen = lambda: print("\033c", end="")
validate_link = lambda link: True if "spotify.com" in link else False
validate_selection = lambda selected_songs: True if len(selected_songs) != 0 else False


def extract_track_links(sp, link):
    id = link.split("/")[-1].split("?")[0]
    if "playlist" in link:
        data = sp.playlist(id)
    elif "album" in link:
        data = sp.album(id)

    total_tracks = data["tracks"]["total"]
    offset, limit = 0, 100
    track_links = []

    while offset < total_tracks:
        if "playlist" in link:
            results = sp.playlist_items(id, offset=offset, limit=limit)
            for track in results["items"]:
                track_links.append(track["track"]["href"])

        elif "album" in link:
            limit = 40
            results = sp.album_tracks(id, offset=offset, limit=limit)
            for track in results["items"]:
                track_links.append(track["href"])

        offset += limit

    return track_links


def add_metadata(data, path):
    audio = ID3(path)
    audio.add(TIT2(encoding=3, text=data["name"]))
    audio.add(
        TPE1(encoding=3, text=",".join([artist["name"] for artist in data["artists"]]))
    )
    audio.add(TALB(encoding=3, text=data["album"]["name"]))
    audio.add(
        APIC(
            encoding=3,
            mime="image/jpeg",
            type=3,
            desc="Cover",
            data=requests.get(data["album"]["images"][0]["url"]).content,
        )
    )

    audio.save(v2_version=3)


def pool_download(
    downloader,
    links,
    custom_hook,
    download_path,
    quality,
    metadata,
    max_concurrent=4,
):
    with concurrent.futures.ThreadPoolExecutor(max_concurrent) as executor:
        tasks = [
            executor.submit(
                downloader.download_track,
                link,
                custom_hook,
                download_path,
                quality,
                metadata,
            )
            for link in links
        ]
        return concurrent.futures.wait(tasks,return_when="FIRST_EXCEPTION")


def progress_hook(self, d, custom_hook=None):
    if d["status"] == "downloading":
        percent = "".join(
            char for char in d["_percent_str"] if char.isdigit() or char == "."
        )
        if self.is_library:
            percent = "".join(
                char for char in d["_percent_str"] if char.isdigit() or char == "."
            )
            self.progress_dict.update(
                {d["info_dict"]["id"]: float(percent[3:]) * 0.943}
            )
            progress = sum(self.progress_dict.values()) / self.total_tracks
            custom_hook(str(round(progress, 1)))
        else:
            custom_hook(str(round(float(percent[3:]) * 0.943, 1)))
    elif d["status"] == "downloaded":
        custom_hook("97.8")
    elif d["status"] == "transfered":
        custom_hook("100")


def parallel_searches(self, track_links):
    with concurrent.futures.ThreadPoolExecutor(max_workers=6) as executor:
        tasks = [executor.submit(self.track_details, link) for link in track_links]
        concurrent.futures.wait(tasks)
    return [task.result() for task in tasks]


def display_settings(settings):
    table = Table(
        title="[bold green]SETTINGS",
        show_header=False,
        row_styles=["yellow", "yellow", "yellow"],
        show_lines=True,
    )
    for i in settings.keys():
        table.add_row(i, settings[i])
    console.print(Panel(table, expand=False, border_style="blue"))


def display_table(dic: dict):
    title = {
        "ALBUM ID": "ALBUM INFO",
        "PLAYLIST ID": "PLAYLIST INFO",
        "TRACK ID": "TRACK INFO",
    }

    table = Table(
        title=f"[bold green]{title[list(dic.keys())[0]]}",
        show_header=False,
        row_styles=["yellow", "yellow", "yellow"],
        show_lines=True,
    )

    for i in dic.keys():
        table.add_row(i, str(dic[i]))
    console.print(Panel(table, expand=False, border_style="blue"))


def display_selected_songs(li):
    table = Table(
        title="[bold green]SELECTED SONGS",
        show_header=False,
        row_styles=["yellow", "yellow", "yellow"],
        show_lines=True,
    )
    table.add_column("Songs")
    for i in li:
        table.add_row(i)
    console.print(Panel(table, expand=False, border_style="blue"))


def get_spotidex_data_directory():
    spotidex_path = Path(os.path.expanduser("~"), ".spotidex")
    spotidex_path.mkdir(parents=True, exist_ok=True)

    return spotidex_path


def get_ffmpeg():
    os_name = platform.system().lower()
    os_arch = platform.machine().lower()

    ffmpeg_url = FFMPEG_URLS.get(os_name, {}).get(os_arch)
    ffmpeg_path = (
        get_spotidex_data_directory()
        / f"ffmpeg{''.join(('.exe' if os_name == 'windows' else ''))}"
    )

    if ffmpeg_path.exists():
        current_path = Path().cwd()
        os.chdir(get_spotidex_data_directory())
        try:
            subprocess.run(
                ["./ffmpeg.exe"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
            )
        except Exception:
            try:
                ffmpeg_binary = requests.get(ffmpeg_url, stream=True, timeout=10)
            except Exception:
                print(
                    "Netowork Error: Please check your internet connction and try later"
                )
                exit()
            with open(ffmpeg_path, "wb") as ffmpeg_file, tqdm(
                total=100,
                unit="%",
                unit_scale=True,
                unit_divisor=1024,
                dynamic_ncols=True,
                bar_format="Initializing : |{bar:50}| {percentage:.0f}% ",
            ) as progress_bar:
                try:
                    for data in ffmpeg_binary.iter_content(chunk_size=1024):
                        ffmpeg_file.write(data)
                        progress_bar.update(
                            len(data)
                            * 100
                            / int(ffmpeg_binary.headers.get("content-length", 0))
                        )
                except Exception:
                    progress_bar.close()
                    print(
                        "Netowork Error: Please check your internet connction and try later"
                    )
                    exit()
            if os_name in ["linux", "darwin"]:
                ffmpeg_path.chmod(ffmpeg_path.stat().st_mode | stat.S_IEXEC)
            os.chdir(current_path)
            return ffmpeg_path
        os.chdir(current_path)
        return ffmpeg_path

    try:
        ffmpeg_binary = requests.get(ffmpeg_url, stream=True, timeout=10)
    except Exception:
        print("Netowork Error: Please check your internet connction and try later")
        exit()

    with open(ffmpeg_path, "wb") as ffmpeg_file, tqdm(
        total=100,
        unit="%",
        unit_scale=True,
        unit_divisor=1024,
        dynamic_ncols=True,
        bar_format="Initializing : |{bar:50}| {percentage:.0f}% ",
    ) as progress_bar:

        try:
            for data in ffmpeg_binary.iter_content(chunk_size=1024):
                ffmpeg_file.write(data)
                progress_bar.update(
                    len(data)
                    * 100
                    / int(ffmpeg_binary.headers.get("content-length", 0))
                )
        except Exception:
            progress_bar.close()
            print("Netowork Error: Please check your internet connction and try later")
            exit()

    if os_name in ["linux", "darwin"]:
        ffmpeg_path.chmod(ffmpeg_path.stat().st_mode | stat.S_IEXEC)

    return ffmpeg_path



from setuptools import setup,find_packages

with open("README.md","r") as f:
    description = f.read()

setup(
    name="spotidex",
    version="0.0.3",
    long_description=description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "mutagen==1.47.0",
        "Requests==2.31.0",
        "rich==13.7.1",
        "spotipy==2.23.0",
        "textual==0.55.1",
        "tqdm==4.66.1",
        "yt_dlp==2024.3.10",
        "ytmusicapi==1.3.2",
    ],
    author="libin-codes",
    author_email="libinlalu000@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "spotidex = spotidex:main",  
        ],
    },
    python_requires=">=3.10",
)

